#include <netlink/genl/genl.h>
#include <netlink/genl/ctrl.h>
#include <netlink/msg.h>
#include <netlink/attr.h>
#include <net/if.h>

#include "device.h"

struct _Device {
        int id;
        int nl80211_id;
        char* name;
        int operation;
        bool powersave;
        struct nl_sock* socket;
        struct nl_msg* message;
        struct nl_cb* callback;
        
};

static Device*
device_free_real(Device* device)
{
        if (device == NULL)
                return NULL;
        if (device->socket != NULL)
                nl_socket_free(device->socket);
        if (device->callback != NULL)
                nl_cb_put(device->callback);
        free(device);
        return NULL;
}

static bool
clean_message(Device* device) 
{
        if (device != NULL && device->message != NULL) {
                nlmsg_free(device->message);
                device->message = NULL;
        }
        return false;
}

static bool
clean_callback(Device* device) 
{
        if (device != NULL && device->callback != NULL) {
                nl_cb_put(device->callback);
                device->callback = NULL;
        }
        return clean_message(device);
}

static int
ack_handler(struct nl_msg* msg, void* arg)
{
        int *ret = arg;
        *ret = 0;
        return NL_STOP;
}

static int
valid_handler(struct nl_msg *msg, void *arg)
{
        Device* device = arg;
        if (device->operation != NL80211_CMD_GET_POWER_SAVE)
                return NL_SKIP;

        device->powersave = false;

	struct nlattr *attrs[NL80211_ATTR_MAX + 1];
	struct genlmsghdr *gnlh = nlmsg_data(nlmsg_hdr(msg));

	nla_parse(attrs, NL80211_ATTR_MAX, genlmsg_attrdata(gnlh, 0),
		  genlmsg_attrlen(gnlh, 0), NULL);

	if (!attrs[NL80211_ATTR_PS_STATE])
		return NL_SKIP;

	if (nla_get_u32(attrs[NL80211_ATTR_PS_STATE]) == 1)
                device->powersave = true;

	return NL_SKIP;
}

Device*
device_new(char* name)
{
        Device* device = calloc(1, sizeof(Device));
        if (device == NULL)
                return NULL;
        device->id = if_nametoindex(name);
        device->socket = nl_socket_alloc();
        if (device->socket == NULL)
                return device_free_real(device);
        if (genl_connect(device->socket))
                return device_free_real(device);
        nl_socket_set_buffer_size(device->socket, 8192, 8192);
        int ret;
        setsockopt(nl_socket_get_fd(device->socket), SOL_NETLINK,
                   NETLINK_EXT_ACK, &ret, sizeof(int));
        device->nl80211_id = genl_ctrl_resolve(device->socket, "nl80211");
        if (device->nl80211_id < 0)
                return device_free_real(device);

        return device;
}

static bool
device_powersave_operation(Device* device,
                           int     operation,
                           bool    value)
{
        device->operation = operation;

        device->message = nlmsg_alloc();
        if (device->message == NULL)
                return false;
        genlmsg_put(device->message, 0, 0, device->nl80211_id,
                    0, 0, operation, 0);
        if (nla_put(device->message, NL80211_ATTR_IFINDEX,
                    sizeof(int), &device->id) < 0)
                return clean_message(device);

        int ps = value ? 1 : 0;
        if (nla_put(device->message, NL80211_ATTR_PS_STATE,
                    sizeof(int), &ps) < 0)
                return clean_message(device);

        device->callback = nl_cb_alloc(NL_CB_DEFAULT);
        if (device->callback == NULL)
                return clean_message(device);

        if (nl_send_auto_complete(device->socket, device->message) < 0)
                return clean_callback(device);

        int err = 1;
        nl_cb_set(device->callback, NL_CB_ACK, NL_CB_CUSTOM,
                  ack_handler, &err);
        nl_cb_set(device->callback, NL_CB_VALID, NL_CB_CUSTOM,
                  valid_handler, device);

        while (err > 0)
		nl_recvmsgs(device->socket, device->callback);

        nl_cb_put(device->callback);
        nlmsg_free(device->message);

        device->callback = NULL;
        device->message = NULL;

        return true;
}

bool
device_get_powersave(Device* device, bool* powersave)
{
        bool r = device_powersave_operation(
                device, NL80211_CMD_GET_POWER_SAVE, false);
        *powersave = device->powersave;
        return r;
}

bool
device_set_powersave(Device* device,
                     bool    powersave)
{
        return device_powersave_operation(
                device, NL80211_CMD_SET_POWER_SAVE, powersave);
}

void
device_free(Device* device)
{
        device_free_real(device);
}
